#ifndef BOND_H
#define BOND_H

#include "spin.h"
#include "rng.h"
#include <cmath>
#include <utility>



uint getRoot(uint l0, std::vector<uint>& g) {
  while(l0 != g[l0-1]) l0 = g[l0-1];
  return l0;
};

uint getRoot(uint l0, uint l1, std::vector<uint>& g) {
  while(l0 != g[l0-1]) l0 = g[l0-1];
  while(l1 != g[l1-1]) l1 = g[l1-1];
  return (l0<l1)?l0:l1;
};

void glue(uint l0, uint l1, std::vector<uint>& g){
  if (l0 != l1) {
    g[l0-1] = getRoot(l0,l1,g);
    g[l1-1] = g[l0-1];
  }
};




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUBBOND %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Subbond {

  void link(ePtr& a,ePtr& b) {
    a->ot = b;
    b->in = a;
  };
  
  void link0(ePtr& a,ePtr& b,ePtr& c,ePtr& d) {
    a = b; 
    c = d;
  };
  // d --> a --> c
  // f --> b --> e
  void link1(ePtr& a, ePtr& b, ePtr& c, ePtr& d, ePtr& e, ePtr& f) {
    a->ot = c; 
    a->in = d; 
    b->ot = e; 
    b->in = f;
  };
  void linkHB(ePtr& e0, ePtr& e1, ePtr& l0, ePtr& l1){
    if (e0->state == 1) {
      if (l0->in == l1) {
	link1(e0,e1,e1,e1,e0,e0);
	link0(l0->in,l1,l1->ot,l0);
      }
      else {
	link1(e0,e1,e1,l0->in,l1->ot,e0);
	link0(l0->in->ot,e0,l1->ot->in,e1);
	link0(l0->in,l1,l1->ot,l0);
      }
    }
    else {
      if (l1->in == l0) {
	link1(e0,e1,e1,e1,e0,e0);
	link0(l0->ot,l1,l1->in,l0);
      }
      else {
	link1(e1,e0,e0,l1->in,l0->ot,e1);
	link0(l0->ot->in,e0,l1->in->ot,e1);
	link0(l0->ot,l1,l1->in,l0);
      }
    }
  };
  void linkDB(ePtr& e0, ePtr& e1, ePtr& l0, ePtr& l1){
    if (e0->state == 1) {
      link1(e0,e1,l1,l0->in,l0,l1->in);
      link0(l0->in,e1,l1->in,e0);
      link0(e0->in->ot,e0,e1->in->ot,e1);
    }
    else {
      link1(e0,e1,l0->ot,l1,l1->ot,l0);
      link0(l0->ot,e1,l1->ot,e0);
      link0(e0->ot->in,e0,e1->ot->in,e1);
    }
  };

  void random_breakup(ePtr& e0,ePtr& e1,ePtr& l0,ePtr& l1) {
    double r = exp_dist();
    if (r < decayHB/(decayHB+decayDB)) {
      // ######### link HB here! ##########
      if (e0->state == 1) {link(e0,e1);link(l0,l1);} 
      else                {link(e1,e0);link(l1,l0);}
    } else {
      // ######### link DB here! ##########
      if (e0->state == 1) {link(e0,l1);link(l0,e1);} 
      else                {link(l1,e0);link(e1,l0);}
    }
  };

 public:
  //-------------------------------------------------------
  // VARIABLES --------------------------------------------
  //-------------------------------------------------------
  Subspin* sp[2];     // pointer to subspins linked
  double Jx;         // x-component of coupling constant
  double Jy;         // y-component of coupling constant
  double Jz;         // z-component of coupling constant
  bool ab;       // is this an alphabond?
  // decay constants:
  double decayHB;    // horizontal breakup
  double decayDB;    // diagonal breakup
  double decayG;     // glue (kind of operator)
  bool do_HB,do_DB,do_G_on_diff,do_G_on_same;
  //-------------------------------------------------------

  //-------------------------------------------------------
  // TOOLS ------------------------------------------------
  //-------------------------------------------------------
  //-------------------------------------------------------

  //-------------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR ----------------------------
  //-------------------------------------------------------
  Subbond(double jx,       // x-comp. of coupling constant
	  double jy,       // y-comp. of coupling constant
	  double jz,       // z-comp. of coupling constant
	  bool _ab,        // alphabond?
	  Subspin* s0,     // subspin linked by this
	  Subspin* s1)     // subspin linked by this
    :Jx(jx),Jy(jy),Jz(jz),ab(_ab) {
    sp[0] = s0;
    sp[1] = s1;
    
    double Jxy = (double)0.5*(Jx+Jy);
    double absJxy = (Jxy>=0.)?Jxy:-Jxy;
    double absJz = (Jz>=0.)?Jz:-Jz;
    
    // XY isotropic
    if (Jx == Jy) {
      // AF region
      if (Jz >= absJxy) {
	decayHB = absJxy*(double)0.5;
	do_HB = true;
	decayDB = (double)0.;
	do_DB = false;
	decayG = (Jz==Jxy)?(double)0.:(Jz-absJxy)*(double)0.5;
	do_G_on_diff = (Jz==Jxy)?false:true;
	do_G_on_same = false;
      }
      
      // FM region
      else if (Jz <= -absJxy) {
	decayHB = (double)0.;
	do_HB = false;
	decayDB = absJxy*(double)0.5;
	do_DB = true;
	decayG = (Jz==Jxy)?(double)0.:(absJz-absJxy)*(double)0.5;
	do_G_on_same = (Jz==Jxy)?false:true;
	do_G_on_diff = false;
      }
      
      // XY-like region
      else if (absJz < absJxy) {
	decayHB = (absJxy + Jz)*(double)0.25;
	do_HB = true;
	decayDB = (absJxy - Jz)*(double)0.25;
	do_DB = true;
	decayG = (double)0.;
	do_G_on_same = false;
	do_G_on_diff = false;
      }
    }
  };
  ~Subbond(){};
  //-------------------------------------------------------

  //-------------------------------------------------------
  // MEMBER FUNCTIONS -------------------------------------
  //-------------------------------------------------------
  int spread_ablabel(double tS){
    int res(0);
    if (!ab) return res;
    double tE;
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter en0(sp[0]->end()),en1(sp[1]->end());
    ssIter em0(sp[0]->end()),em1(sp[1]->end());
    em0--;em1--;
    IntervalController interval(lc0,lc1,en0,en1,tE,speq);
    if ((*lc0)->t == (*lc1)->t && lc0 != em0) {
      (*lc0)->ab = ab;
      (*lc1)->ab = ab;
      res++;
    }
    while (true) {
      if (interval.next(tS,tE,speq,lc0,lc1)) {
	if ((*lc0)->t == (*lc1)->t && lc0 != em0) {
	  (*lc0)->ab = ab;
	  (*lc1)->ab = ab;
	  res++;
	}
      }
      else break;
    }
    return res;
  };
  bool linking_possible(ePtr& e0,ePtr& e1,ePtr& l0,ePtr& l1) {
    if (e0->state == e1->state) return false;
    if (e0->state == 1) {
      if (e0->ot != NULL) return false;
      if (l0->ot != NULL) return false;
      if (e1->in != NULL) return false;
      if (l1->in != NULL) return false;
    } else {
      if (e0->in != NULL) return false;
      if (l0->in != NULL) return false;
      if (e1->ot != NULL) return false;
      if (l1->ot != NULL) return false;
    }
    return true;
  };
  int updateOP(double tS){
    // header
    int res(0);
    
    double tE;
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter en0(sp[0]->end()),en1(sp[1]->end());
    ssIter em0(sp[0]->end()),em1(sp[1]->end());
    em0--;em1--;
    IntervalController interval(lc0,lc1,en0,en1,tE,speq);
    ePtr e0,e1,l0,l1;
    double r;
    int type;
    
    // in this version of relinking it is possible
    // that two events of the same time link different
    // subspins of neighbouring spins, this leads to an
    // error in relinking
    // at every relink it must be checked if it is possible
    // to relink and if the edges are yet unlinked
    // this is costly, because the large majority of
    // edges do not need to be checked anyway!!!
    if ((*lc0)->t == (*lc1)->t && lc0 != em0) {
      e0 = *lc0;e1 = *lc1;
      lc0++;lc1++;
      l0 = *lc0;l1 = *lc1;
      lc0--;lc1--;
      if (linking_possible(e0,e1,l0,l1)) random_breakup(e0,e1,l0,l1);
    }
    while (true) {
      if (interval.next(tS,tE,speq,lc0,lc1)) {
	if ((*lc0)->t == (*lc1)->t && lc0 != em0) {
	  e0 = *lc0;e1 = *lc1;
	  lc0++;lc1++;
	  l0 = *lc0;l1 = *lc1;
	  lc0--;lc1--;
	  if (linking_possible(e0,e1,l0,l1)) random_breakup(e0,e1,l0,l1);
	}
      }
      else break;
    }
  };

  int spreadOP(double tS, EdgeCollector* edcol){
    int res(0);
    double tE;
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    IntervalController interval(lc0,lc1,e0,e1,tE,speq);
    if (!speq && do_HB) res += spread(0,tS,tE,lc0,lc1,edcol);
    if ( speq && do_DB) res += spread(1,tS,tE,lc0,lc1,edcol);
    while (true) {
      if (interval.next(tS,tE,speq,lc0,lc1)) {
	if (!speq && do_HB) res += spread(0,tS,tE,lc0,lc1,edcol);
	if ( speq && do_DB) res += spread(1,tS,tE,lc0,lc1,edcol);
      }
      else break;
    }
    return res;
  };
  int spreadOP_help(double tS, EdgeCollector* edcol){
    int res(0);
    double tE;
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    IntervalController interval(lc0,lc1,e0,e1,tE,speq);
    if (!speq && do_HB) res += spread_help(0,tS,tE,lc0,lc1,edcol);
    if ( speq && do_DB) res += spread_help(1,tS,tE,lc0,lc1,edcol);
    while (true) {
      if (interval.next(tS,tE,speq,lc0,lc1)) {
	if (!speq && do_HB) res += spread_help(0,tS,tE,lc0,lc1,edcol);
	if ( speq && do_DB) res += spread_help(1,tS,tE,lc0,lc1,edcol);
      }
      else break;
    }
    return res;
  };
  
  int spread(int ty, double tS,double tE,ssIter lc0,ssIter lc1,EdgeCollector* edcol) {
    int res(0);
    double decay;
    switch (ty) {
    case 0: decay = decayHB; break;
    case 1: decay = decayDB; break;
    }
    while ((tS += exp_dist(decay)) < tE) {
      ePtr e0 = edcol->get(tS,(*lc0)->state);
      ePtr e1 = edcol->get(tS,(*lc1)->state);

      switch (ty) {
      case 0: linkHB(e0,e1,(*lc0),(*lc1)); break;
      case 1: linkDB(e0,e1,(*lc0),(*lc1)); break;
      }
      // tell the edges if they are alphabonds
      e0->ab = ab;
      e1->ab = ab;
      sp[0]->insert(lc0,e0);
      sp[1]->insert(lc1,e1);
      res++;
    }
    return res;
  };
  int spread_help(int ty, double tS,double tE,ssIter lc0,ssIter lc1,EdgeCollector* edcol) {
    show("************************************************");
    show("spreading between line " << sp[0]->line << " and " << sp[1]->line);
    std::cout << std::setprecision(16);
    show("interval from " << tS << " to " << tE);
    int res(0);
    double decay;
    switch (ty) {
    case 0: decay = decayHB; break;
    case 1: decay = decayDB; break;
    }
    double tt = exp_dist(decay);
    while ((tS += tt) < tE) {
      show("   waiting time from exp_dist() = " << tt);
      show("new breakup inserted at " << tS);
      show("                type is " << ty);
      ePtr e0 = edcol->get(tS,(*lc0)->state);
      ePtr e1 = edcol->get(tS,(*lc1)->state);
      
      switch (ty) {
      case 0: linkHB(e0,e1,(*lc0),(*lc1)); break;
      case 1: linkDB(e0,e1,(*lc0),(*lc1)); break;
      }
      // tell the edges if they are alphabonds
      e0->ab = ab;
      e1->ab = ab;
      sp[0]->insert(lc0,e0);
      sp[1]->insert(lc1,e1);
      res++;
      tt = exp_dist(decay);
    }
    
    std::cout << std::setprecision(10);
    show("************************************************");
    return res;
  };

  int spreadG(double tS, std::vector<uint>& g){
    uint res(0); // not used yet!!!!
    double tE;
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    IntervalController interval(lc0,lc1,e0,e1,tE,speq);
    if (!speq && do_G_on_diff && ((tS+=exp_dist(decayG))<tE)) glue((*lc0)->label,(*lc1)->label,g);
    if ( speq && do_G_on_same && ((tS+=exp_dist(decayG))<tE)) glue((*lc0)->label,(*lc1)->label,g);
    while (true) {
      if (interval.next(tS,tE,speq,lc0,lc1)) {
	if (!speq && do_G_on_diff && ((tS+=exp_dist(decayG))<tE)) glue((*lc0)->label,(*lc1)->label,g);
	if ( speq && do_G_on_same && ((tS+=exp_dist(decayG))<tE)) glue((*lc0)->label,(*lc1)->label,g);
      } else break;
    }
    return res; // not used yet!!!!
  };

  double diagenergy(double tS) {
    if (Jz == 0.) return 0.;
    double res((double)0.),tE;
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    IntervalController interval(lc0,lc1,e0,e1,tE,speq);
    res += (speq)?tE-tS:tS-tE;
    while (true) {
      if (interval.next(tS,tE,speq,lc0,lc1)) res+=(speq)?tE-tS:tS-tE;
      else break;
    }
    return res*Jz/(double)4.;
  };

  double numbonds(double tS) {
    double tE;
    uint res(0);
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    IntervalController interval(lc0,lc1,e0,e1,tE,speq,res);
    while(true) if(!interval.next(tS,tE,speq,lc0,lc1,res)) break;
    return (double)res;
  };

  void get_operators(vec_d &times,vec_s &types) {
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    IntervalController interval(lc0,lc1,e0,e1,times,types);
    while(true) if(!interval.next(lc0,lc1,times,types)) break;
  };
  
  void get_spatialWindNum(vec_i &wN) {
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    ePtr beg0;
    if (sp[0]->front()->state == 1) beg0 = sp[0]->front()->in;
    else beg0 = sp[0]->front()->ot;
    IntervalController interval(lc0,lc1,e0,e1,wN,beg0);
    while(true) if(!interval.next(lc0,lc1,wN)) break;
  };

  void get_intervals(uint bd,std::map<double,short>& intervals,double tS) {
    double tE;
    uint res(0);
    bool speq;
    ssIter lc0(sp[0]->begin()),lc1(sp[1]->begin());
    ssIter e0(sp[0]->end()),e1(sp[1]->end());
    IntervalController interval(lc0,lc1,e0,e1,tE,speq,res);

    if (res == 1) {
      intervals[tE] = ((*lc0)->state==1)?bd+1:-bd-1;
      res = 0;
    }
    while(true) {
      if(!interval.next(tS,tE,speq,lc0,lc1,res)) break;
      if (res == 1) {
	intervals[tE] = ((*lc0)->state==1)?bd+1:-bd-1;
	res = 0;
      }
    }
  };

  //-------------------------------------------------------
  // FUNCTIONS FOR DISPLAY --------------------------------
  //-------------------------------------------------------
  //-------------------------------------------------------
  // display: ---------------------------------------------
  //-------------------------------------------------------
  void display() {
    show("   " << sp[0]->line << " <--> " << 
	 sp[1]->line << ":");
    showl("   " << "Jx = " << Jx);
    showl(";  " << "Jy = " << Jy);
    showl(";  " << "Jz = " << Jz);
    showl(";  " << "ab = " << ab );
    showl(";  " << "decayHB = " << decayHB);
    showl(";  " << "decayDB = " << decayDB);
    showl(";  " << "decayG  = " << decayG );
    show (";  " << "do_HB = " << do_HB);
    showl("   " << "do_DB = " << do_DB);
    showl(";  " << "do_G_on_diff = " << do_G_on_diff);
    show (";  " << "do_G_on_same = " << do_G_on_same);
  };
  //-------------------------------------------------------
  // displayEntries: --------------------------------------
  //-------------------------------------------------------
  void displayEntries() {
/*     if (!empty()) { */
/*       iterator iter = --end(); */
/*       for (;iter!=begin();iter--) (*iter)->display(); */
/*       (*iter)->display(); */
/*     } */
  };
  //-------------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// BOND %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Bond : public std::vector<Subbond*> {
 public:
  //-------------------------------------------------------
  // VARIABLES --------------------------------------------
  //-------------------------------------------------------
  Spin* Sp[2];
  //-------------------------------------------------------

  //-------------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR ----------------------------
  //-------------------------------------------------------
  Bond(double jx,double jy,double jz,bool ab,Spin* S0,Spin* S1) {
    Sp[0] = S0;
    Sp[1] = S1;

    for (uint i=0;i<S0->size();i++) 
      for (uint j=0;j<S1->size();j++) push_back(new Subbond(jx,jy,jz,ab,(*S0)[i],(*S1)[j]));
  };
  ~Bond(){for(iterator iter=begin();iter!=end();iter++) delete *iter;};
  //-------------------------------------------------------

  //-------------------------------------------------------
  // MEMBER FUNCTIONS -------------------------------------
  //-------------------------------------------------------
  double diagenergy(double tS) {
    double res((double)0.);
    for (uint i=0;i<size();i++) res += this[0][i]->diagenergy(tS);
    return res;
  };
  double numbonds(double tS) {
    double res((double)0.);
    for (uint i=0;i<size();i++) res += this[0][i]->numbonds(tS);
    return res;
  };
  void get_intervals(uint bd,std::map<double,short>& intervals,double tS) {
    for (uint i=0;i<size();i++) this[0][i]->get_intervals(bd,intervals,tS);
  };
  void get_spatialWindNum(vec_i &wN) {
    for (uint i=0;i<size();i++) this[0][i]->get_spatialWindNum(wN);
  };

  //-------------------------------------------------------

  //-------------------------------------------------------
  // FUNCTIONS FOR DISPLAY --------------------------------
  //-------------------------------------------------------
  void display(){
    show("bond between spin at (" << 
	 Sp[0]->row << "," << 
	 Sp[0]->col << ") and (" << 
	 Sp[1]->row << "," <<
	 Sp[1]->col << "):");
    for (uint i=0;i<size();i++) this[0][i]->display();
  };
  //-------------------------------------------------------
  
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



#endif // BOND_H ///:~
