/**
 * \file demo_ored.cpp
 *
 * \brief Demonstrates an application using the Simulation class.
 *
 */

#include "simulation.h"

using namespace std;

int main(int argc, char **argv) {
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // INPUT %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  if (argc <= 7) {
    show("Not enough input arguments!!!");
    show("1...Model");
    show("2...Lx");
    show("3...beta");
    show("4...alpha");
    show("5...Delta");
    show("6...Nmeas");
    show("7...filebase");
    exit(0);
  }

  // the input parameters are given to the Simulation class below
  string     sModel         (argv[1]);	// defines the model, a chain here!
  int        Lx        (atoi(argv[2]));	// length of the chain
  double     beta      (atof(argv[3]));	// inverse temperature = length of imaginary time
  double     alpha     (atof(argv[4]));	// bond alternation
  double     Delta     (atof(argv[5]));	// XXZ exchange anisotropy
  int        Nmeas     (atoi(argv[6]));	// number of measurements
  string     base           (argv[7]);	// string to be used as base for filenames

  string parmfile(base + ".prm");	// name of file that contains further parameters
  string measfile(base);		// basename of file that contains measurements
					// base.uni ... unimproved standard estimators
					// base.imp ... improved estimators
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  // SIMULATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  Simulation S(sModel,Lx,beta,alpha,Delta,Nmeas,base); // this constructor works yet for 1D models only!!!
  S.read_parameter(parmfile);
  S.init_spinsystem();
  for (uint i=0;i<S.SA->spin.size();i++) S.SA->spin[i]->choose_random_state();
  S.init_uni_estimators(measfile);
  S.init_imp_estimators(measfile);
  // look for an existing state, read it and continue the run in the current directory
  // statefiles are: state of the spin chain, AND the random number generator
  if (S.SA->read_state()) {
    S.num -= S.Ntherm;
    S.Ntherm = 0;
    showlog("Ntherm set to 0");
  } else {
    showlog("No state found, chosing random spin state at tau=0.");
    for (uint i=0;i<S.SA->spin.size();i++) S.SA->spin[i]->choose_random_state();
  }
  // check if measurements are defined, otherwise define default measurements for demo
  // this can be used to prevent accidental runs on computer clusters without output
  if (!S.check_measurements()) {
    showlog("No measurements found in files " << measfile 
	    << "*!!! \nMeasuring some default demo observables!");
    S.init_uni_estimator("E");
    S.init_imp_estimator("imp_stsus");
    S.init_imp_estimator("imp_stsuspm");
  }
  S.simulate();
  S.finish();
  //%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



  return 0;
}
