#ifndef EDGE_H
#define EDGE_H

#include "types.h"
#include "macros.h"
#include "rng.h"
#include <list>




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// EDGE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Edge;
typedef Edge* ePtr;
class Edge {
 public:
  //-------------------------------------------------------
  // VARIABLES --------------------------------------------
  //-------------------------------------------------------
  double    t;
  ePtr      in,ot;
  short     state;
  uint      label;
  bool      ab;     // part of an alphabond?
  //-------------------------------------------------------

  //-------------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR ----------------------------
  //-------------------------------------------------------
  Edge():t((double)0.),in(NULL),ot(NULL),state(0),label(0),ab(false){};
  Edge(double tt):t(tt),in(NULL),ot(NULL),state(0),label(0),ab(false){};
  Edge(double tt,uint s):t(tt),in(NULL),ot(NULL),state(s),label(0),ab(false){};
  Edge(double tt,uint s,bool):t(tt),state(s),label(0),ab(false) {
    in = &this[0];
    ot = &this[0];
  };
  ~Edge(){};
  //-------------------------------------------------------
  // MEMBER FUNCTIONS -------------------------------------
  //-------------------------------------------------------
  int check_ends() {
    if (in != NULL && in->ot == NULL) in->ot = &this[0];
    if (ot != NULL && ot->in == NULL) ot->in = &this[0];
    if (in == NULL || ot == NULL) return 1;
    else return 0;
  };
  bool check_inot()	 {return (in->ot==&this[0] && ot->in==&this[0]);};
  bool check_zerolinks() {return (ot != NULL && in != NULL);};

  void choose_random_state() {state=(exp_dist()<0.5)?-1:1;};
  ePtr pushLoop(uint l){label = l;return ot;};
  void flipclear(){flip();clear();};
  void flip(){
    state = -state;
    ePtr temp(in);
    in = ot;
    ot = temp;
  };
  void clear(){label = 0;};
  //-------------------------------------------------------

  //-------------------------------------------------------
  // FUNCTIONS FOR DISPLAY --------------------------------
  //-------------------------------------------------------
  void display(){
    std::cout << std::setprecision(16);
    showl("      " << &this[0] << " at " << 
	  std::setw(10) << t << ": " << 
	  std::setw(3) << state << ", " << 
	  std::setw(3) << ab << ", " << 
	  std::setw(5) << label << ";\n");
    if (in == NULL)
      showl("\t\t\tin = NULL\n");
    else
      showl("\t\t\tin = " << in << " at " << std::setw(20) << std::fixed << in->t << "\n");
    if (ot == NULL)
      show("\t\t\tot = NULL\n");
    else
      show("\t\t\tot = " << ot << " at " << std::setw(20) << std::fixed << ot->t);
    std::cout << std::setprecision(10);
  };
  void display(short line){
    std::cout << std::setprecision(16);
    showl(std::setw(4) << line << ": " << &this[0] << " at " << 
	  std::setw(10) << t << ": " << 
	  std::setw(3) << state << ", " << 
	  std::setw(3) << ab << ", " << 
	  std::setw(5) << label << ";\n");
    if (in == NULL)
      showl("\t\t\tin = NULL\n");
    else
      showl("\t\t\tin = " << in << " at " << std::setw(20) << std::fixed << in->t << "\n");
    if (ot == NULL)
      show("\t\t\tot = NULL\n");
    else
      show("\t\t\tot = " << ot << " at " << std::setw(20) << std::fixed << ot->t);
    std::cout << std::setprecision(10);
 };
  void shortdisplay(){
    show("      " << &this[0] << " at " << 
	 std::setw(10) << t << ": " << 
	 std::setw(3) << state << ", " << 
	 std::setw(5) << label << ";\n" <<
	 "\tin = " << std::setw(8) << in << "\n" <<
	 "\tot = " << std::setw(8) << ot);
  };
  //-------------------------------------------------------
};
typedef std::list<ePtr> eCont;
typedef std::list<ePtr>::iterator eContIter;
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



#endif // EDGE_H ///:~
