#ifndef FUNCTIONS_H
#define FUNCTIONS_H

#include "bond.h"




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// BASECLASS FOR FUNCTION OBJECTS %%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Function {
 public:
  //-----------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR --------------------------
  //-----------------------------------------------------
  Function(){};
  ~Function(){};
  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  virtual int operator()(Spin*){return 0;};
  virtual int operator()(Spin&){return 0;};
  virtual int operator()(Spin*,int){return 0;};
  virtual int operator()(Spin&,int){return 0;};
  virtual int operator()(Subspin*){return 0;};
  virtual int operator()(Subspin&){return 0;};
  virtual int operator()(Subspin*,int){return 0;};
  virtual int operator()(Subspin&,int){return 0;};
  virtual int operator()(Bond*){return 0;};
  virtual int operator()(Bond&){return 0;};
  virtual int operator()(Bond*,int){return 0;};
  virtual int operator()(Bond&,int){return 0;};
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SPREADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Spreader : public Function {
 public:
  int naO;
  //-----------------------------------------------------
  // TOOLS ----------------------------------------------
  //-----------------------------------------------------
  EdgeCollector* edcol;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  int operator()(Bond& b){return this[0](&b);};
  int operator()(Bond *b){
    int res(0);
    for (uint i=0;i<b->size();i++) {
      int temp((*b)[i]->spreadOP((double)0.,edcol));
      res += temp;
      if ((*b)[i]->ab) naO += temp;
    }
    return res;
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// HELPSPREADER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class HelpSpreader : public Function {
 public:
  int naO;
  //-----------------------------------------------------
  // TOOLS ----------------------------------------------
  //-----------------------------------------------------
  EdgeCollector* edcol;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  int operator()(Bond& b){return this[0](&b);};
  int operator()(Bond *b){
    int res(0);
    for (uint i=0;i<b->size();i++) {
      int temp((*b)[i]->spreadOP_help((double)0.,edcol));
      res += temp;
      if ((*b)[i]->ab) naO += temp;
    }
    return res;
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// UPDATER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Updater : public Function {
 public:
  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  int operator()(Bond& b){return this[0](&b);};
  int operator()(Bond *b){
    int res(0);
    for (uint i=0;i<b->size();i++) (*b)[i]->updateOP((double)0.);
    return res;
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// LOOPER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Looper : public Function {
 public:
  //-----------------------------------------------------
  // VARIABLES ------------------------------------------
  //-----------------------------------------------------
  ePtr head;
  ssIter iter;
  uint label;
  bool prepared;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  Looper():label(1),prepared(true){};
  ~Looper(){};
  void clear() {label = 1;};

  virtual int  operator()(Spin& s){return this[0](&s);};
  virtual int  operator()(Spin *s) {
    if (prepared) {
      for (uint i=0;i<s->size();i++) {
	iter = (*s)[i]->begin();
	for(;iter!=(*s)[i]->end();iter++) {
	  head = *iter;
	  if (head->label == 0) {
	    while (head->label == 0) head = head->pushLoop(label);
	    label++;
	  }
	}
      }
    }
    else {
      for (uint i=0;i<s->size();i++) {
	iter = (*s)[i]->begin();
	head = *iter;
	if (head->label == 0) {
	  while (head->label == 0) head = head->pushLoop(label);
	  label++;
	}
      }
    }
    return 0;
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// GLUER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Gluer : public Function {
 public:
  //-----------------------------------------------------
  // VARIABLES ------------------------------------------
  //-----------------------------------------------------
  std::vector<uint> glue;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  void init(uint l) {
    for (uint i=0;i<l-1;i++) glue.push_back(i+1);
  };
  void clear() {glue.clear();};
  void gluedown() {for (uint i=0;i<glue.size();i++) glue[i] = getRoot(i+1,glue);};
  int operator()(Bond& s){return this[0](&s);};
  int operator()(Bond *s){
    for (uint i=0;i<s->size();i++) 
      if ((*s)[i]->do_G_on_diff || (*s)[i]->do_G_on_same) (*s)[i]->spreadG((double)0.,glue);
    return 0;
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// FLIPPER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Flipper : public Function {
 public:

  //-----------------------------------------------------
  // TOOLS ----------------------------------------------
  //-----------------------------------------------------
  std::vector<bool> flip;
  ssIter iter;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  virtual void decide(int num) {
    for (int i=0;i<num-1;i++) flip.push_back((exp_dist()>0.5));
  };
  virtual void decide(std::vector<uint>& g) {
    for (uint i=0;i<g.size();i++) {
      if (g[i] == i+1) flip.push_back((exp_dist()>0.5));
      else flip.push_back(flip[g[i]-1]);
    }
  };

  void set(uint fl,std::vector<uint>& g) {
    show(g.size());
    for (uint i=0;i<g.size();i++) {
      if (g[i] == i+1) {
	if (i == fl-1) flip.push_back(true);
	else flip.push_back(false);
      }
      else flip.push_back(flip[g[i]-1]);
    }
  };
  void clear() {flip.clear();};
  int operator()(Spin& s){return this[0](&s);};
  int operator()(Spin *s){
    for (uint i=0;i<s->size();i++) {
      iter = (*s)[i]->begin();
      if (flip[(*iter)->label-1]) (*iter)->flipclear();
      else (*iter)->clear();
      iter++;
      while (iter != (*s)[i]->end()) {
	if (flip[(*iter)->label-1]) (*iter)->flipclear();
	else (*iter)->clear();
	iter++;
      }
    }
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// REMOVER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Remover : public Function {
 public:
  int naO;
  //-----------------------------------------------------
  // VARIABLES & TOOLS ----------------------------------
  //-----------------------------------------------------
  ssIter iter,before;
  EdgeCollector* edcol;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  int operator()(Spin& s){return this[0](&s);};
  int operator()(Spin *s){
    int res(0);
    for (uint i=0;i<s->size();i++) {
      iter = (*s)[i]->begin();
      before = iter;
      iter++;
      while (iter != (*s)[i]->end()) {
 	if ((*iter)->state == (*before)->state) {
	  if ((*before)->ab) naO++;
	  (*s)[i]->cleanremove(before,iter,edcol);
	  res++;
	} 
	before = iter;
	iter++;
      }
    }
    return res;
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// OPENREMOVER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class OpenRemover : public Function {
 public:
  int naO;
  //-----------------------------------------------------
  // VARIABLES & TOOLS ----------------------------------
  //-----------------------------------------------------
  ssIter iter,before;
  EdgeCollector* edcol;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  int operator()(Spin& s){return this[0](&s);};
  int operator()(Spin *s){
    int res(0);
    for (uint i=0;i<s->size();i++) {
      iter = (*s)[i]->begin();
      before = iter;
      iter++;
      while (iter != (*s)[i]->end()) {
	(*before)->in = NULL;
	(*before)->ot = NULL;
 	if ((*iter)->state == (*before)->state) {
	  if ((*before)->ab) naO++;
	  (*s)[i]->remove(before,edcol);
	  res++;
	} 
	before = iter;
	iter++;
      }
      (*before)->in = NULL;
      (*before)->ot = NULL;
    }
    return res;
  };
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SWAPPER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Swapper : public Function {
 public:
  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  int operator()(Spin& s){return this[0](&s);};
  int operator()(Spin *s){s->swap();return 0;};
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SWAPCONNECTOR %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class SwapConnector : public Function {
 public:
  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  int operator()(Spin& s){return this[0](&s);};
  int operator()(Spin *s){s->swapconnect();return 0;};
  //-----------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// DISPLAYER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Displayer : public Function {
 public:
  // for subspins
  int operator()(Spin *s){
    s->display_withEntries();
/*     s->display(); */
/*     s->displayEntries(); */
    return 0;
  };
  int operator()(Spin& s){return this[0](&s);};
  int operator()(Spin *s,int i){
    s->display();
    s->displayEntries(i);
    return 0;
  };
  int operator()(Spin& s,int i){return this[0](&s,i);};

  // for subbonds
  int operator()(Bond *s){s->display();return 0;};
  int operator()(Bond& s){return this[0](&s);};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// CHECKER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Checker : public Function {
 public:
  std::string check;

  // for spins
  int operator()(Spin& s){return this[0](&s);};
  int operator()(Spin *s){
    int res(0);
    if (check.compare("boundary")==0) return s->check_boundary();
    return 0;
  };

  // for subpins
  int operator()(Subspin& s){return this[0](&s);};
  int operator()(Subspin *s){
    int res(0);
    if (check.compare("inot")==0) return s->check_inot();
    if (check.compare("zerolinks")==0) return s->check_zerolinks();
    return 0;
  };

  // for subbonds
  int operator()(Bond& b){return this[0](&b);};
  int operator()(Bond *b){return 0;};
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




#endif // FUNCTIONS_H ///:~
