#ifndef RNG_H
#define RNG_H

#include <cmath>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <vector>
#include <ctime>




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// STANDARD RNG %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class RNG {
  std::string name;
  unsigned long imax;

  virtual unsigned long iengine(){return rand();};

 public:
  unsigned long calls;
  std::vector<unsigned long> seeds;

  RNG():name("RAND"),imax(RAND_MAX),calls(0),seeds(1,1){};
  RNG(bool sowing):name("RAND"),imax(RAND_MAX),calls(0),seeds(1,1){
    if (sowing) seeds[0] = sow();
    srand(seeds[0]);
  };
  RNG(std::string n):name("RAND"),imax(RAND_MAX),calls(0),seeds(1,1){};
  RNG(std::string n, bool sowing):name("RAND"),imax(RAND_MAX),calls(0),seeds(1,1){
    if (sowing) seeds[0] = sow();
    srand(seeds[0]);
  };
/*   RNG(string n):name(n),imax(RAND_MAX),calls(0),tmp(1),seeds(1,1){}; */
/*   RNG(string n, bool sowing):name(n),imax(RAND_MAX),calls(0),tmp(1),seeds(1,1){ */
/*     if (sowing) seeds[0] = sow(); */
/*     srand(seeds[0]); */
/*   }; */
  
  void set_imax(unsigned long im) {imax = im;};
  unsigned long get_imax() {return imax;};
  std::string get_name() {return name;};

  virtual double operator()(){return (double)iengine()/(double)imax;};

  virtual unsigned long sow(){return time(NULL);};

  virtual void write_state() {
    std::string name=get_name()+"_state";
    std::ofstream out(name.c_str());
    std::vector<unsigned long> res;
    get_state(res);
    for (int i=0;i<res.size();i++) out << res[i] << std::endl;
    out.close();
  };
  virtual void read_state() {
    std::clog << "reading state of " << get_name() << std::endl;
    std::string name=get_name()+"_state";
    std::ifstream in(name.c_str());
    std::vector<unsigned long> res(2);
    if(in.is_open()) {
      in >> res[0];
      in >> res[1];
      set_state(res);
      in.close();
    } else std::cerr << "infile " << name << "not found!!!" << std::endl;
  };
  virtual void get_state(std::vector<unsigned long>& res) {
    res.clear();
    res.push_back(seeds[0]);
    res.push_back(calls);
  };
  virtual void set_state(std::vector<unsigned long> res) {
    if (res.size()>0) seeds[0] = res[0];
    srand(seeds[0]);
    if (res.size()>1) for (int i=0;i<res[1];i++) iengine();
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// EXPONENTIAL DISTRIBUTION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class ExpDist {
 public:
  RNG *rng;

  ExpDist(){rng = new RNG;};
  ExpDist(std::string rng_name){load(rng_name);};
  ~ExpDist(){delete rng;};

  void load(std::string rng_name){load(rng_name,false);};
  void load(std::string rng_name, bool sowing){rng = new RNG(sowing);};

  double operator()(){return((*rng)());};
  double operator()(double decay){return(1.e-15 - 1./decay*log((*rng)()));};

  void write_state() {rng->write_state();};
  void  read_state() {rng-> read_state();};

};
ExpDist exp_dist;
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#endif // RNG_H ///:~
