#ifndef SPIN_H
#define SPIN_H

#include "edge.h"
#include <queue>



//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// EDGECOLLECTOR %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class EdgeCollector : public std::queue<ePtr> {
  int N;
 public:
  EdgeCollector():N(0){};
  ~EdgeCollector(){
    while(!empty()) {
      delete front();
      pop();
    }
  };

  ePtr get(double t, short s) {
    if (empty()) {
      N++;
      return new Edge(t,s);
    }
    else {
      ePtr result = front();
      if (result->label != 0) {
	show(result->label << " labelled edge " << result << " given out by edcol!!!");
	show("  time before = " << result->t);
	exit(0);
      }
      pop();
      result->t = t;
      result->state = s;
      result->in = NULL;
      result->ot = NULL;
      return result;
    }
  };
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SUBSPIN %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Subspin : public eCont {
 public:
  //-------------------------------------------------------
  // VARIABLES --------------------------------------------
  //-------------------------------------------------------
  short line;
  //-------------------------------------------------------

  //-------------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR ----------------------------
  //-------------------------------------------------------
  Subspin(short l, short s, double t):line(l){insert(end(),new Edge(t,s,true));};
  ~Subspin(){for(iterator iter=begin();iter!=end();iter++) delete *iter;};
  //-------------------------------------------------------

  //-------------------------------------------------------
  // MEMBER FUNCTIONS -------------------------------------
  //-------------------------------------------------------
  int check_inot() {
    int res(0);
    for (iterator iter=begin();iter!=end();iter++) {
      if (!(*iter)->check_inot()) res++;
    }
    return res;
  };
  int check_inot(double tS, double tE) {
    int res(0);
    iterator iter=begin();
    while ((*iter)->t < tS) iter++;
    while ((*iter)->t < tE) {
      if (!(*iter)->check_inot()) res++;
      iter++;
    }
    return res;
  };
  int check_zerolinks() {
    int res(0);
    for (iterator iter=begin();iter!=end();iter++) {
      if (!(*iter)->check_zerolinks()) res++;
    }
    return res;
  };
  int check_zerolinks(double tS, double tE) {
    int res(0);
    iterator iter=begin();
    while ((*iter)->t < tS) iter++;
    while ((*iter)->t < tE) {
      if (!(*iter)->check_zerolinks()) res++;
      iter++;
    }
    return res;
  };

  void choose_random_state(){front()->choose_random_state();};

  void get_edges(vec_d &times,vec_s &states) {
    iterator iter(begin());
    while (iter != end()) {
      times.push_back((*iter)->t);
      states.push_back((*iter)->state);
      iter++;
    }
  };

  void resetremove(iterator tbrem,EdgeCollector *edcol) {
    (*tbrem)->in = NULL;
    (*tbrem)->ot = NULL;
    edcol->push(*tbrem);
    erase(tbrem);
  };
  void remove(iterator tbrem,EdgeCollector *edcol) {
    edcol->push(*tbrem);
    erase(tbrem);
  };
  void cleanremove(iterator tbrem, iterator after,EdgeCollector *edcol) {
    if ((*tbrem)->ot != (*tbrem)->in) {
      if ((*tbrem)->state == 1) {
	(*after)->in = (*tbrem)->in;
	(*tbrem)->in->ot = (*after);
      }
      else {
	(*after)->ot = (*tbrem)->ot;
	(*tbrem)->ot->in = (*after);
      }
    }
    
    edcol->push(*tbrem);
    erase(tbrem);
  };

  short state() {return front()->state;};
  uint getLabelAt(double t) {
    iterator iter(begin());
    while ((*iter)->t < t) iter++;
    return (*iter)->label;
  };
  void disconnect_boundaries() {
    if (size() > 1) {
      iterator last = end();
      last--;
      // disconnect at boundaries
      if ((*last)->ot != NULL) {
	(*last)->ot->in = NULL;
	(*last)->ot = NULL;
      }
      if ((*last)->in != NULL) {
	(*last)->in->ot = NULL;
	(*last)->in = NULL;
      }
    }
  };
  //-------------------------------------------------------

  //-------------------------------------------------------
  // FUNCTIONS FOR DISPLAY --------------------------------
  //-------------------------------------------------------
  void display(){
    show("   " << line << ": " << 
	 back()->t << ": state " << 
	 front()->state << "(" <<
	 back()->state  << "); label " << 
	 front()->label << "(" << 
	 back()->label  << "); " << 
	 size());
  };
  void displayEntries() {for (iterator iter(begin()) ;iter!=end();iter++) (*iter)->display(line);};
  void displayEntries(int i) {
    for (iterator iter(begin()) ;iter!=end();iter++) {
      if (i==1 && (*iter)->ab) (*iter)->display();
      else if (!(*iter)->ab) (*iter)->display();
    }
  };
  //-------------------------------------------------------
};
typedef Subspin::iterator ssIter;
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SPIN %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Spin : public std::vector<Subspin*> {
 public:
  //-------------------------------------------------------
  // VARIABLES --------------------------------------------
  //-------------------------------------------------------
  short   row;      // y-position in 2D array
  short   col;      // x-position in 2D array
  double  t0;       // start time, usually 0.
  double  t1;       // end time, usually beta
  bool    stag;
  //-------------------------------------------------------

  //-------------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR ----------------------------
  //-------------------------------------------------------
  Spin(short& l,
       short  n,    // number of subspins
       short  r,    // y-position in 2D array
       short  c,    // x-position in 2D array
       short  s,    // state
       double tt0,  // usually 0.
       double tt1,  // usually beta
       bool   st):row(r),col(c),t0(tt0),t1(tt1),stag(st) {
    if (s>0 &&  (s%2)!=(n%2)) s=(n%2);
    if (s<0 && -(s%2)!=(n%2)) s=-(n%2);
    if (s>n)  s = n;
    if (s<-n) s = -n;
    
    int sbs;
    for (int i=0;i<n;i++) {
      sbs = (s>0)? 1:-1;
      push_back(new Subspin(l,sbs,t1));
      s -= sbs;
      l++;
    }
  };
  ~Spin(){for(iterator iter=begin();iter!=end();iter++) delete *iter;};
  //-------------------------------------------------------

  //-------------------------------------------------------
  // MEMBER FUNCTIONS -------------------------------------
  //-------------------------------------------------------
  bool check_boundary() {
    int state0(0),stateb(0);
    for(iterator iter=begin();iter!=end();iter++) {
      state0+=(*iter)->front()->state;
      stateb+=(*iter)->back()->state;
    }
    if (state0 != stateb) return 1;
    else return 0;
  };

  void choose_random_state() {for (uint i=0;i<size();i++) this[0][i]->choose_random_state();};

  short state() {
    short res(0);
    for (uint i=0;i<size();i++) res += this[0][i]->state();
    return res;
  };

  void swapconnect() {
    reconnect_boundaries();
    swap();
  };
  void swap() {
    Subspin *si,*sj;
    if (size() > 1) {
      for (uint i=0;i<size();i++) {
	for (uint j=i+1;j<size();j++) {
	  si = this[0][i];
	  sj = this[0][j];
	  if (si->front()->state == sj->front()->state) {
	    if (exp_dist() < 0.5) { //swap the edge connections
	      swap(si,sj);
	    }
	  }
	}
      }
    }
  };
  void swap (Subspin *si, Subspin *sj) {
    ePtr iF(si->front()),jF(sj->front());
    ePtr temp;

    if (iF->state == 1) {
      temp = iF->in;
      iF->in = jF->in;
      jF->in = temp;
      iF->in->ot = iF;
      jF->in->ot = jF;
    }
    else {
      temp = iF->ot;
      iF->ot = jF->ot;
      jF->ot = temp;
      iF->ot->in = iF;
      jF->ot->in = jF;
    }
  };
  short frontline() {return front()->line;};
  short backline()  {return back()->line;};
  short nextline()  {return front()->line+size();};
  short prevline()  {return back()->line-size()+1;};
  
  void disconnect_boundaries() {for (uint i=0;i<size();i++) this[0][i]->disconnect_boundaries();};
  void reconnect_boundaries() {
    std::vector<ePtr> up_front,dn_front,up_back,dn_back;
    ePtr e;
    for (uint i=0;i<size();i++) {
      e = this[0][i]->front();
      if (e->state == 1) up_front.push_back(e);
      else dn_front.push_back(e);
      e = this[0][i]->back();
      if (e->state == 1) up_back.push_back(e);
      else dn_back.push_back(e);
    }
    
    if (up_front.size() != up_back.size()) {show("NOT THE SAME NUBMER OF UP SPINS AT FRONT AND BACK!!!");exit(0);}
    if (dn_front.size() != dn_back.size()) {show("NOT THE SAME NUBMER OF DOWN SPINS AT FRONT AND BACK!!!");exit(0);}
    
    for (uint i=0;i<up_front.size();i++) {
      up_front[i]->in = up_back[i];
      up_back[i]->ot = up_front[i];
    }
    for (uint i=0;i<dn_front.size();i++) {
      dn_front[i]->ot = dn_back[i];
      dn_back[i]->in = dn_front[i];
    }
  };
  
  void get_position(short line, double &r, double &c) {
    for (uint i=0;i<size();i++) {
      if (line == this[0][i]->line) {
	r = row + i*0.1;
	c = col + i*0.1;
	break;
      }
    }
  }
  //-------------------------------------------------------

  //-------------------------------------------------------
  // FUNCTIONS FOR DISPLAY --------------------------------
  //-------------------------------------------------------
  void display(){
    show(row << "," << col << "; " <<
	 t0 << " - " << t1 << "; stag(" << stag << 
	 "); " << size() << ":");
    for (uint i=0;i<size();i++) this[0][i]->display();
  }
  void displayEntries(){
    for (uint i=0;i<size();i++) 
      this[0][i]->displayEntries();
  }
  void displayEntries(int ii){
    for (uint i=0;i<size();i++) 
      this[0][i]->displayEntries(ii);
  }
  void display_withEntries(){
    show(row << "," << col << "; " <<
	 t0 << " - " << t1 << "; stag(" << stag << 
	 "); " << size() << ":");
    for (uint i=0;i<size();i++) {
      this[0][i]->display();
      this[0][i]->displayEntries();
    }
  };
  //-------------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// INTERVAL CONTROLLER %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class IntervalController {
  //-------------------------------------------------------
  // VARIABLES --------------------------------------------
  //-------------------------------------------------------
  uint labove,lbelow;
  short sabove,sbelow;
  ssIter s0,s1;
  ssIter end0,end1;
  ePtr beg0;
  double     currentTStart;
  //-------------------------------------------------------

 public:
  //-------------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR ----------------------------
  //-------------------------------------------------------
  IntervalController(ssIter i0, ssIter i1, ssIter e0, ssIter e1, double& tE)
    :s0(i0),s1(i1),end0(e0),end1(e1){
    if ((*s0)->t == (*s1)->t) {
      tE = (*s0)->t;
      s0++;s1++;
    }
    else if ((*s0)->t < (*s1)->t) {
      tE = (*s0)->t;
      s0++;
    }
    else {
      tE = (*s1)->t;
      s1++;
    }
    currentTStart = tE;
  };
  IntervalController(ssIter i0, ssIter i1, ssIter e0, ssIter e1, double& tE, bool& speq)
    :s0(i0),s1(i1),end0(e0),end1(e1){
    
    speq = ((*s0)->state == (*s1)->state);

    if ((*s0)->t == (*s1)->t) {
      tE = (*s0)->t;
      s0++;
      s1++;
    }
    else if ((*s0)->t < (*s1)->t) {
      tE = (*s0)->t;
      s0++;
    }
    else {
      tE = (*s1)->t;
      s1++;
    }
    currentTStart = tE;
  };
  IntervalController(ssIter i0, ssIter i1, ssIter e0, ssIter e1, double& tE, bool& speq, uint& nb)
    :s0(i0),s1(i1),end0(e0),end1(e1){
    
    speq = ((*s0)->state == (*s1)->state);

    if ((*s0)->t == (*s1)->t) {
      tE = (*s0)->t;
      nb++;
      s0++;
      s1++;
    }
    else if ((*s0)->t < (*s1)->t) {
      tE = (*s0)->t;
      s0++;
    }
    else {
      tE = (*s1)->t;
      s1++;
    }
    currentTStart = tE;
  };
  IntervalController(ssIter i0, ssIter i1, ssIter e0, ssIter e1, vec_i& wN, ePtr& b0)
    :s0(i0),s1(i1),end0(e0),end1(e1),beg0(b0){
    
    if ((*s0)->t == (*s1)->t) eval_wN(wN);
    else if ((*s0)->t < (*s1)->t) s0++;
    else s1++;
  };
  IntervalController(ssIter i0, 
		     ssIter i1, 
		     ssIter e0, 
		     ssIter e1,
		     vec_d &times,
		     vec_s &types)
    :s0(i0),s1(i1),end0(e0),end1(e1){
    
    if ((*s0)->t == (*s1)->t) {
      times.push_back((*s0)->t);
      if ((*s0)->state == 1) {
	if ((*s0)->ot == (*s1)) {
	  types.push_back(0); // HB
	}
	else {
	  types.push_back(1); // DB
	}
      } else {
	if ((*s0)->in == (*s1)) {
	  types.push_back(0); // HB
	}
	else {
	  types.push_back(1); // DB
	}
      }
      s0++;s1++;
    }
    else if ((*s0)->t < (*s1)->t) s0++;
    else s1++;
  };
  ~IntervalController(){};
  //-------------------------------------------------------

  //-------------------------------------------------------
  // MEMBER FUNCTIONS -------------------------------------
  //-------------------------------------------------------
  bool next(double& t1,ssIter& loc0,ssIter& loc1) {
    if (s0 == end0 && s1 == end1) return false;
    else {
      loc0 = s0;
      loc1 = s1;
      
      if ((*s0)->t == (*s1)->t) {t1 = (*s0)->t;s0++;s1++;}
      else if ((*s0)->t < (*s1)->t) {t1 = (*s0)->t;s0++;}
      else {t1 = (*s1)->t;s1++;}
      return true;
    }
  };
  bool next_operator(double& t1,ssIter& loc0,ssIter& loc1) {
    if (s0 == end0 && s1 == end1) return false;
    else {
      if ((*s0)->t == (*s1)->t) {
	loc0 = s0;
	loc1 = s1;
	t1 = (*s0)->t;
	s0++;
	s1++;
      }
      else {
	while (next(t1,loc0,loc1)) {
	  if ((*s0)->t == (*s1)->t) break;
	}
      }
      return true;
    }
  };
  bool next(double& t0,double& t1,ssIter& loc0,ssIter& loc1) {
    if (s0 == end0 && s1 == end1) return false;
    else {
      t0 = currentTStart;
      loc0 = s0;
      loc1 = s1;
      
      if ((*s0)->t == (*s1)->t) {t1 = (*s0)->t;s0++;s1++;}
      else if ((*s0)->t < (*s1)->t) {t1 = (*s0)->t;s0++;}
      else {t1 = (*s1)->t;s1++;}
      currentTStart = t1;
      return true;
    }
  };
  bool next(double& t0, double& t1, bool& speq, ssIter& loc0, ssIter& loc1) {
    if (s0 == end0 && s1 == end1) return false;
    else {
      t0 = currentTStart;
      speq = ((*s0)->state == (*s1)->state);
      loc0 = s0;
      loc1 = s1;
      
      if ((*s0)->t == (*s1)->t) {t1 = (*s0)->t;s0++;s1++;}
      else if ((*s0)->t < (*s1)->t) {t1 = (*s0)->t;s0++;}
      else {t1 = (*s1)->t;s1++;}
      currentTStart = t1;
      return true;
    }
  };
  bool next(double& t0, double& t1, bool& speq, ssIter& loc0, ssIter& loc1,uint& nb) {
    if (s0 == end0 && s1 == end1) {nb--; return false;}
    else {
      t0 = currentTStart;
      speq = ((*s0)->state == (*s1)->state);
      loc0 = s0;
      loc1 = s1;
      
      if ((*s0)->t == (*s1)->t) {
	t1 = (*s0)->t;
	nb++;
	s0++;
	s1++;
      }
      else if ((*s0)->t < (*s1)->t) {t1 = (*s0)->t;s0++;}
      else {t1 = (*s1)->t;s1++;}
      currentTStart = t1;
      return true;
    }
  };
  bool next(ssIter& loc0, ssIter& loc1,vec_i &wN) {
    if (s0 == end0 && s1 == end1) return false;
    else {
      loc0 = s0;
      loc1 = s1;
      
      if ((*s0)->t == (*s1)->t) eval_wN(wN);
      else if ((*s0)->t < (*s1)->t) s0++;
      else s1++;
      return true;
    }
  };
  bool next(ssIter& loc0, ssIter& loc1,vec_d &times,vec_s &types) {
    if (s0 == end0 && s1 == end1) return false;
    else {
      loc0 = s0;
      loc1 = s1;
      
      if ((*s0)->t == (*s1)->t) {
	times.push_back((*s0)->t);
	if ((*s0)->state == 1) {
	  if ((*s0)->ot == (*s1)) 
	    types.push_back(0); // HB
	  else {
	    types.push_back(1); // DB
	  }
	} else {
	  if ((*s0)->in == (*s1)) 
	    types.push_back(0); // HB
	  else {
	    types.push_back(1); // DB
	  }
	}
	s0++;s1++;
      }
      else if ((*s0)->t < (*s1)->t) s0++;
      else s1++;
      return true;
    }
  };

  void eval_wN(vec_i &wN) {
    lbelow = (*s0)->label;
    sbelow = (*s0)->state;
    s0++;s1++;
    if (s0 == end0) {
      labove = beg0->label;
      sabove = beg0->state;
    } else {
      labove = (*s0)->label;
      sabove = (*s0)->state;
    }
    wN[lbelow] += sbelow;
    wN[labove] -= sabove;
  };
  //-------------------------------------------------------
};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




#endif // SPIN_H ///:~
