#ifndef SPINSYSTEM_H
#define SPINSYSTEM_H

#include "basecell.h"
#include "functions.h"
#include <string>
#include <fstream>
#include <sstream>




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// AUXILIARY CLASS FOR FILE HANDLING %%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class myFile {
  std::string pipe;

 public:
  myFile():pipe("temp_pipe"){};
  ~myFile(){};
  
  bool sec_open(std::ifstream &in,std::string name) {return sec_open(in,name,true);};
  bool sec_open(std::ifstream &in,std::string name,bool strict) {
    if (name.substr(name.size()-4,4).compare(".bz2")==0) name = name.substr(0,name.size()-4);
    showlogl(name);
    in.open(name.c_str());
    if(!in.is_open()) {
      std::string name_bz2(name+".bz2");
      in.open(name_bz2.c_str());
      if(!in.is_open()) {
	if (strict) {showlog(".bz2 infile not found! Exiting now ..."); exit(0);}
	else showlog(".bz2 infile not found!"); 
      }
      else {
	in.close();
	showlog(".bz2");
	std::string exec = "if test -e " + pipe + "; then cd .; else mkfifo " + pipe + "; fi;";
	system(exec.c_str());
	exec = "bunzip2 -c " + name_bz2 + " > " + pipe + " &";
	system(exec.c_str());
	in.open(pipe.c_str());
      }
    }
    else showlog("");
    return in.is_open();
  };
  
  void sec_close(std::ifstream &in) {
    in.close();
    std::string exec("rm -f " + pipe);
    system(exec.c_str());
  };

};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// SPINSYSTEM %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class Spinsystem {
 public:
  //-----------------------------------------------------
  // VARIABLES & MODEL PARAMETERS -----------------------
  //-----------------------------------------------------
  std::string strModel;
  short  Lx,Ly;   // system size
  double t0,beta,alpha,Delta;
  bool   withPro; // projections on?
  int    nO,naO;  // number of operators
  //-----------------------------------------------------
  // SPINS AND BONDS ------------------------------------
  //-----------------------------------------------------
  std::vector<Spin*>    spin; // pointer to Spins
  std::vector<Spin*>::iterator sp;
  std::map<short,Subspin*> subspin; // pointer to Subspins
  std::map<short,Subspin*>::iterator sb;
  std::vector<Bond*> bond; // pointer to Bonds   
  std::vector<Bond*>::iterator bo;
  std::vector<short> deleted; // lineID of del. spins
  std::vector<Subbond*> subbond; // pointer to Subbonds
  //-----------------------------------------------------

  //-----------------------------------------------------
  // TOOLS ----------------------------------------------
  //-----------------------------------------------------
  EdgeCollector* edcol;
  Spreader spread;
  Looper   loop;
  Gluer    glue;
  Flipper  flip;
  Remover  remove;
  Swapper  swap;
  // new:
  Updater       updateOP;
  OpenRemover   openremove;
  SwapConnector swapconnect;
  HelpSpreader helpspread;
  std::map<double,short> intervals;
  //-----------------------------------------------------

  //-----------------------------------------------------
  // CONSTRUCTORS & DESTRUCTOR --------------------------
  //-----------------------------------------------------
  Spinsystem():t0((double)0.),nO(0),naO(0){};
  Spinsystem(const char* fin_name)
    :t0((double)0.),nO(0),naO(0){init(fin_name);};
  Spinsystem(short lx,     // # of spins in x-dir
	     short ly,     // # of spins in y-dir
	     mat_i& S,     // spinsize in basecell
	     mat_d& xjx,   // see "basecell.h" for
	     mat_d& xjy,   // further information
	     mat_d& xjz,
	     mat_d& yjx,mat_d& yjy,mat_d& yjz, 
	     mat_b& f,     // spin staggered?
	     mat_b& xA,    // x-bond with alpha?
	     mat_b& yA,    // y-bond with alpha? 
	     double t1,   // usually 0.
	     double t2    // usually beta 
	     ):nO(0),naO(0){
    // calls init
    init(lx,ly,S,xjx,xjy,xjz,yjx,yjy,yjz,f,xA,yA,t1,t2);
  };
  Spinsystem   (short     lx,    // # basecells in x-dir
		short     ly,    // # basecells in y-dir
		Basecell& bc,    // basecell to reprodce
		double   t1,    // usually 0.
		double   t2     // usually beta
		):nO(0),naO(0){
    // calls init with parameters from basecell
    init(lx,ly,bc,t1,t2);
  };
  ~Spinsystem(){
    for (uint i=0;i<bond.size();i++) delete bond[i];
    for (uint i=0;i<spin.size();i++) delete spin[i];
    delete edcol;
  };
  //-----------------------------------------------------
  // MEMBER FUNCTIONS -----------------------------------
  //-----------------------------------------------------
  virtual void init(short lx, short ly, mat_i& S, 
	    mat_d& xjx, mat_d& xjy, mat_d& xjz, 
	    mat_d& yjx, mat_d& yjy, mat_d& yjz,
	    mat_b& f, mat_b& xA, mat_b& yA,
	    double t1, double t2){
    edcol = new EdgeCollector;
    spread.edcol = edcol;
    remove.edcol = edcol;
    openremove.edcol = edcol;
    helpspread.edcol = edcol;

    Ly = ly;
    Lx = lx;

    //---------------------------------------------------
    // we initialise the spins --------------------------
    //---------------------------------------------------
    short nx(S[0].size()),ny(S.size()),n,s,l(0);
    bool st;
    std::map<int,std::map<int,Spin*> > sm;
    for (short row=0;row<Ly;row++) {
      for (short col=0;col<Lx;col++) {
	n = S[row%ny][col%nx];
	st = f[row%ny][col%nx];
	s = ((row+col)%2 == 0)?n:-n;
	spin.push_back(new Spin(l,n,row,col,s,t1,t2,st));
	sm[row][col] = spin.back();
	for (int i=0;i<n;i++) {
	  Subspin* sub((*sm[row][col])[i]);
	  subspin[sub->line] = sub;
	}
      }
    }
    
    //---------------------------------------------------
    // we initialise the bonds, for each spin we must
    // create two bonds: one to the right neighbor and
    //                   one to the upper neighbor
    // if Ly = 1 no upper neighbors must be considered
    //---------------------------------------------------
    short xnx(xjx[0].size()),xny(xjx.size());
    short ynx(yjx[0].size()),yny(yjx.size());
    short row,col;
    double x,y,z;
    sp = spin.begin();
    for (;sp!=spin.end();sp++) {
      row = (*sp)->row;
      col = (*sp)->col;
      x = xjx[row%xny][col%xnx];
      y = xjy[row%xny][col%xnx];
      z = xjz[row%xny][col%xnx];
      if (x != 0. || y != 0. || z != 0.) {
	bond.push_back(new Bond(x,y,z,xA[row%xny][col%xnx],*sp,sm[row][(col+1)%Lx]));
	for (short i=0;i<bond.back()->size();i++) subbond.push_back((*bond.back())[i]);
      }
      x = yjx[row%yny][col%ynx];
      y = yjy[row%yny][col%ynx];
      z = yjz[row%yny][col%ynx];
      if (x != 0. || y != 0. || z != 0.) {
	bond.push_back(new Bond(x,y,z,yA[row%yny][col%ynx],*sp,sm[(row+1)%Ly][col]));
	for (short i=0;i<bond.back()->size();i++) subbond.push_back((*bond.back())[i]);
      }
    }
  };
  void init(short lx,short ly,Basecell& bc,double t1,double t2) {
    edcol = new EdgeCollector;
    spread.edcol = edcol;
    remove.edcol = edcol;
    openremove.edcol = edcol;
    helpspread.edcol = edcol;

    strModel= bc.translate(bc.model);
    t0 = t1;
    beta = t2;
    alpha = bc.alpha;
    Delta = bc.Delta;
    init(lx * bc.S[0].size(),ly * bc.S.size(),
	 bc.S,bc.xJx,bc.xJy,bc.xJz,bc.yJx,bc.yJy,bc.yJz,
	 bc.f,bc.xA,bc.yA,
	 t1,t2);
  };
  void init(std::string){
    edcol = new EdgeCollector;
    spread.edcol = edcol;
    remove.edcol = edcol;
    openremove.edcol = edcol;
    helpspread.edcol = edcol;
  };
  //-----------------------------------------------------
  // forEachSpin/Bond: ----------------------------------
  //-----------------------------------------------------
  int forEachSubspin(Function& f) {
    int res(0);
    sb = subspin.begin();
    for (;sb!=subspin.end();sb++) res += f(sb->second);
    return res;
  };
  int forEachSubspin(Function& f, int ii) {
    int res(0);
    sb = subspin.begin();
    for (;sb!=subspin.end();sb++) res += f(sb->second,ii);
    return res;
  };
  int forEachSpin(Function& f) {
    int res(0);
    for (uint i=0;i<spin.size();i++) res += f(spin[i]);
    return res;
  };
  int forEachSpin(Function& f, int ii) {
    int res(0);
    for (uint i=0;i<spin.size();i++) res += f(spin[i],ii);
    return res;
  };
  int forEachBond(Function& f) {
    int res(0);
    for (uint i=0;i<bond.size();i++) res += f(bond[i]);
    return res;
  };
  int forEachBond(Function& f, int ii) {
    int res(0);
    for (uint i=0;i<bond.size();i++) res += f(bond[i],ii);
    return res;
  };
  //-----------------------------------------------------
  // decorate (with loops): -----------------------------
  //-----------------------------------------------------
  void decorate(){
    // all edges are disconnected after update !!!
    forEachSpin(swapconnect); // this is new!!!
    forEachBond(updateOP);    // this is new!!!
    spread.naO = 0;
    nO  += forEachBond(spread);
    naO += spread.naO;
    loop.prepared = false;
    forEachSpin(loop);
    loop.prepared = true;
    forEachSpin(loop);
    glue.init(loop.label);
    forEachBond(glue);
    glue.gluedown();
  };
  //-----------------------------------------------------
  // update (flip loops): -------------------------------
  //-----------------------------------------------------
  void update() {
    flip.decide(glue.glue);
    forEachSpin(flip);
    // old version:
/*     remove.naO = 0; */
/*     nO  -= forEachSpin(remove)/2; */
/*     naO -= remove.naO/2; */
    // new version:
    openremove.naO = 0;
    nO  -= forEachSpin(openremove)/2;
    naO -= openremove.naO/2;
    loop.clear();
    flip.clear();
    glue.clear();
    intervals.clear();
  };
  //-----------------------------------------------------
  // intervals for integrated measurement: --------------
  //-----------------------------------------------------
  void get_intervals() {
    if (intervals.empty()) for (uint i=0;i<bond.size();i++) bond[i]->get_intervals(i,intervals,t0);
  };

  //-----------------------------------------------------
  // reading from configuration to files: ---------------
  //-----------------------------------------------------
  bool read_state() {
    myFile file;
    std::ifstream tau0state,edges,ran;
    if (!file.sec_open(tau0state,"tau0state.cfg",false)) return false;
    if (!file.sec_open(edges,"edges.cfg",false)) return false;
    if (!file.sec_open(ran,exp_dist.rng->get_name()+"_state",false)) return false;

    file.sec_close(ran);
    exp_dist.rng->read_state();

    short state;
    std::map<uint,std::map<uint,short> > current_state;
    for (uint i=0;i<spin.size();i++) {
      for (uint j=0;j<spin[i]->size();j++) {
	ssIter iter((*spin[i])[j]->begin());
	(*iter)->in = NULL;
	(*iter)->ot = NULL;
	tau0state >> state;
	current_state[i][j] = state;
	(*iter)->state = state;
      }
    }

    int sp,subsp;
    double t;
    while(true) {
      edges >> sp;
      edges >> subsp;
      edges >> t;
      if (edges.eof()) break;
      ssIter last((*spin[sp])[subsp]->end());
      last--;
      (*spin[sp])[subsp]->insert(last,edcol->get(t,current_state[sp][subsp]));
      current_state[sp][subsp] *= -1;
      nO++;
    }

    for (uint i=0;i<spin.size();i++) {
      for (uint j=0;j<spin[i]->size();j++) {
	ssIter last((*spin[i])[j]->end());
	last--;
	(*last)->state = current_state[i][j];
      }
    }

    file.sec_close(tau0state);
    file.sec_close(edges);

    nO /= 2;
    for (uint i=0;i<subbond.size();i++) naO += subbond[i]->spread_ablabel(t0);

    return true;
  };
  //-----------------------------------------------------
  // writing configuration to files: --------------------
  //-----------------------------------------------------
  void write_state() {
    std::ofstream out("tau0state.cfg");
    for (uint i=0;i<spin.size();i++) {
      for (uint j=0;j<spin[i]->size();j++) {
	ssIter iter((*spin[i])[j]->begin());
	showo((*iter)->state);
      }
    }
    out.close();    
    out.open("edges.cfg");
    for (uint i=0;i<spin.size();i++) {
      for (uint j=0;j<spin[i]->size();j++) {
	ssIter iter((*spin[i])[j]->begin());
	double tS((*iter)->t);
	iter++;
	while (iter!=(*spin[i])[j]->end()) {
	  showo(i << " " << j << " " << std::setprecision(16) << tS);
	  tS = (*iter)->t;
	  iter++;
	}
      }
    }
    out.close();    
    exp_dist.write_state();
  };
  //-----------------------------------------------------

};
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





#endif // SPINSYSTEM_H ///:~

